@@ -10,19 +10,23 @@ A simple python lib allowing hierarchical config files in YAML syntax.
 License
 -------
 
-(c) 2014 Klaus Zerwes zero-sys.net
+(c) 2014 - 2020 Klaus Zerwes zero-sys.net
 This package is free software.
 This software is licensed under the terms of the
-GNU LESSER GENERAL PUBLIC LICENSE version 3 or later,
+GNU GENERAL PUBLIC LICENSE version 3 or later,
 as published by the Free Software Foundation.
-See https://www.gnu.org/licenses/lgpl.html
+See https://www.gnu.org/licenses/gpl.html
 """
 
 
 import sys
 import os
 import yaml
+from yaml import parser
 import logging
+from distutils.util import strtobool
+import re
+import io
 from jinja2 import Environment, Undefined, DebugUndefined, StrictUndefined, TemplateError
 
 from . import odyldo
@@ -61,9 +65,11 @@ listTypes = (list, tuple)
 # you may set this to something suitable for you
 jinja2env = Environment(undefined=Undefined)
 
-METHODS = { 'METHOD_SIMPLE':0x0001, 'METHOD_MERGE':0x0002 }
+METHODS = { 'METHOD_SIMPLE':0x0001, 'METHOD_MERGE':0x0002, 'METHOD_SUBSTITUTE':0x0003, }
 METHOD_SIMPLE = METHODS['METHOD_SIMPLE']
 METHOD_MERGE = METHODS['METHOD_MERGE']
+METHOD_SUBSTITUTE = METHODS['METHOD_SUBSTITUTE']
+
 
 class HiYaPyCo():
     """Main class"""
@@ -71,9 +77,12 @@ class HiYaPyCo():
         """
         args: YAMLfile(s)
         kwargs:
-          * method: one of hiyapyco.METHOD_SIMPLE | hiyapyco.METHOD_MERGE
+          * method: one of hiyapyco.METHOD_SIMPLE | hiyapyco.METHOD_MERGE | hiyapyco.METHOD_SUBSTITUTE
+          * mergelists: boolean (default: True) try to merge lists (only makes sense if hiyapyco.METHOD_MERGE or hiyapyco.METHOD_SUBSTITUTE)
           * interpolate: boolean (default: False)
+          * castinterpolated: boolean (default: False) try to cast values after interpolating
           * usedefaultyamlloader: boolean (default: False)
+          * encoding: (default: 'utf-8') encoding used to read yaml files
           * loglevel: one of  the valid levels from the logging module
           * failonmissingfiles: boolean (default: True)
           * loglevelmissingfiles
@@ -86,7 +95,7 @@ class HiYaPyCo():
 
         self.method = None
         if 'method' in kwargs:
-            logging.debug('parse kwarg method: %s ...' % kwargs['method'])
+            logger.debug('parse kwarg method: %s ...' % kwargs['method'])
             if kwargs['method'] not in METHODS.values():
                 raise HiYaPyCoInvocationException(
                         'undefined method used, must be one of: %s' %
@@ -97,7 +106,18 @@ class HiYaPyCo():
         if self.method == None:
             self.method = METHOD_SIMPLE
 
+        self.mergelists = True
+        if 'mergelists' in kwargs:
+            if not isinstance(kwargs['mergelists'], bool):
+                raise HiYaPyCoInvocationException(
+                        'value of "mergelists" must be boolean (got: "%s" as %s)' %
+                        (kwargs['mergelists'], type(kwargs['mergelists']),)
+                        )
+            self.mergelists = kwargs['mergelists']
+            del kwargs['mergelists']
+
         self.interpolate = False
+        self.castinterpolated = False
         if 'interpolate' in kwargs:
             if not isinstance(kwargs['interpolate'], bool):
                 raise HiYaPyCoInvocationException(
@@ -106,6 +126,14 @@ class HiYaPyCo():
                         )
             self.interpolate = kwargs['interpolate']
             del kwargs['interpolate']
+            if 'castinterpolated' in kwargs:
+                if not isinstance(kwargs['castinterpolated'], bool):
+                    raise HiYaPyCoInvocationException(
+                            'value of "castinterpolated" must be boolean (got: "%s" as %s)' %
+                            (kwargs['castinterpolated'], type(kwargs['castinterpolated']),)
+                        )
+                self.castinterpolated = kwargs['castinterpolated']
+                del kwargs['castinterpolated']
 
         if 'usedefaultyamlloader' in kwargs:
             if not isinstance(kwargs['usedefaultyamlloader'], bool):
@@ -127,17 +155,24 @@ class HiYaPyCo():
             self.failonmissingfiles = bool(kwargs['failonmissingfiles'])
             del kwargs['failonmissingfiles']
 
-        self.loglevelonmissingfiles = logging.ERROR
-        if not self.failonmissingfiles:
-            self.loglevelonmissingfiles = logging.WARN
         if 'loglevelmissingfiles' in kwargs:
             logging.getLogger('testlevellogger').setLevel(kwargs['loglevelmissingfiles'])
+            self.loglevelonmissingfiles = logging.getLogger('testlevellogger').getEffectiveLevel()
             del kwargs['loglevelmissingfiles']
+        else:
+            self.loglevelonmissingfiles = logging.ERROR
+            if not self.failonmissingfiles:
+                self.loglevelonmissingfiles = logging.WARN
 
         if 'loglevel' in kwargs:
             logger.setLevel(kwargs['loglevel'])
             del kwargs['loglevel']
 
+        self.encoding = 'utf-8'
+        if 'encoding' in kwargs:
+            self.encoding = kwargs['encoding']
+            del kwargs['encoding']
+
         if kwargs:
             raise HiYaPyCoInvocationException('undefined keywords: %s' % ' '.join(kwargs.keys()))
 
@@ -147,34 +182,57 @@ class HiYaPyCo():
         for arg in args:
             self._updatefiles(arg)
 
-        for yamlfile in self._files:
+        for yamlfile in self._files[:]:
             logger.debug('yamlfile: %s ...' % yamlfile)
-            if '\n' in yamlfile:
-                logger.debug('loading yaml doc from str ...')
-                f = yamlfile
-            else:
-                fn = yamlfile
-                if not os.path.isabs(yamlfile):
-                    fn = os.path.join(os.getcwd(), yamlfile)
-                    logger.debug('path extended for yamlfile: %s' % fn)
-                try:
-                    f = open(fn, 'r')
-                except IOError as e:
-                    logger.log(self.loglevelonmissingfiles, e)
-                    if not fn == yamlfile:
-                        logger.log(self.loglevelonmissingfiles,
-                                'file not found: %s (%s)' % (yamlfile, fn,))
-                    else:
-                        logger.log(self.loglevelonmissingfiles,
-                                'file not found: %s' % yamlfile)
-                    if self.failonmissingfiles:
-                        raise HiYaPyCoInvocationException(
-                                'yaml file not found: \'%s\'' % yamlfile
-                            )
-            if _usedefaultyamlloader:
-                ydata = yaml.safe_load(f)
-            else:
-                ydata = odyldo.safe_load(f)
+            try:
+                if '\n' in yamlfile:
+                    logger.debug('loading yaml doc from str ...')
+                    f = yamlfile
+                    self._load_data(_usedefaultyamlloader, yamlfile)
+                else:
+                    fn = yamlfile
+                    if not os.path.isabs(yamlfile):
+                        fn = os.path.join(os.getcwd(), yamlfile)
+                        logger.debug('path extended for yamlfile: %s' % fn)
+                    try:
+                        with io.open(fn, 'r', encoding=self.encoding) as f:
+                            logger.debug('open4reading: file %s' % f)
+                            self._load_data(_usedefaultyamlloader, f)
+                    except IOError as e:
+                        logger.log(self.loglevelonmissingfiles, e)
+                        if not fn == yamlfile:
+                            logger.log(self.loglevelonmissingfiles,
+                                    'file not found: %s (%s)' % (yamlfile, fn,))
+                        else:
+                            logger.log(self.loglevelonmissingfiles,
+                                    'file not found: %s' % yamlfile)
+                        if self.failonmissingfiles:
+                            raise HiYaPyCoInvocationException(
+                                    'yaml file not found: \'%s\'' % yamlfile
+                                )
+                        self._files.remove(yamlfile)
+                        continue
+            except yaml.parser.ParserError as e:
+                logger.log(self.loglevelonmissingfiles, e)
+                logger.log(self.loglevelonmissingfiles,
+                        'error while parsing yaml %s' % f)
+                if self.failonmissingfiles:
+                    raise HiYaPyCoInvocationException(
+                            'error while parsing file: \'%s\'' % f
+                        )
+                self._files.remove(yamlfile)
+                continue
+
+
+        if self.interpolate:
+            self._data = self._interpolate(self._data)
+
+    def _load_data(self, _usedefaultyamlloader, f):
+        if _usedefaultyamlloader:
+            ydata_generator = yaml.safe_load_all(f)
+        else:
+            ydata_generator = odyldo.safe_load_all(f)
+        for ydata in ydata_generator:
             if logger.isEnabledFor(logging.DEBUG):
                 logger.debug('yaml data: %s' % ydata)
             if self._data is None:
@@ -182,13 +240,14 @@ class HiYaPyCo():
             else:
                 if self.method == METHOD_SIMPLE:
                     self._data = self._simplemerge(self._data, ydata)
-                else:
+                elif self.method == METHOD_MERGE:
                     self._data = self._deepmerge(self._data, ydata)
+                elif self.method == METHOD_SUBSTITUTE:
+                    self._data = self._substmerge(self._data, ydata)
+                else:
+                    raise HiYaPyCoInvocationException('unknown merge method \'%s\'' % self.method)
                 logger.debug('merged data: %s' % self._data)
 
-        if self.interpolate:
-            self._data = self._interpolate(self._data)
-
     def _updatefiles(self, arg):
         if isinstance(arg, strTypes):
             if arg in self._files:
@@ -223,9 +282,24 @@ class HiYaPyCo():
         try:
             si = jinja2env.from_string(s).render(self._data)
         except TemplateError as e:
+            # FIXME: this seems to be broken for unicode str?
             raise HiYaPyCoImplementationException('error interpolating string "%s" : %s' % (s, e,))
         if not s == si:
-            logger.debug('interpolated "%s" to "%s"' % (s, si,))
+            if self.castinterpolated:
+                if not re.match( r'^\d+\.*\d*$', si):
+                    try:
+                        si = bool(strtobool(si))
+                    except ValueError:
+                        pass
+                else:
+                    try:
+                        if '.' in si:
+                            si = float(si)
+                        else:
+                            si = int(si)
+                    except ValueError:
+                        pass
+            logger.debug('interpolated "%s" to "%s" (type: %s)' % (s, si, type(si),))
         return si
 
     def _simplemerge(self, a, b):
@@ -262,6 +336,50 @@ class HiYaPyCo():
                     )
         return a
 
+    def _substmerge(self, a, b):
+        logger.debug('>' * 30)
+        logger.debug('deepmerge %s and %s' % (a, b,))
+        # FIXME: make None usage configurable
+        if b is None:
+            logger.debug('pass as b is None')
+            pass
+
+        # treat listTypes as primitiveTypes in merge
+        # subsititues list, don't merge them
+
+        if a is None or isinstance(b, primitiveTypes) or isinstance(b, listTypes):
+            logger.debug('deepmerge: replace a "%s"  w/ b "%s"' % (a, b,))
+            a = b
+
+        elif isinstance(a, dict):
+            if isinstance(b, dict):
+                logger.debug('deepmerge: dict ... "%s" and "%s"' % (a, b,))
+                for k in b:
+                    if k in a:
+                        logger.debug('deepmerge dict: loop for key "%s": "%s" and "%s"' % (k, a[k], b[k],))
+                        a[k] = self._deepmerge(a[k], b[k])
+                    else:
+                        logger.debug('deepmerge dict: set key %s' % k)
+                        a[k] = b[k]
+            elif isinstance(b, listTypes):
+                logger.debug('deepmerge: dict <- list ... "%s" <- "%s"' % (a, b,))
+                for bd in b:
+                    if isinstance(bd, dict):
+                        a = self._deepmerge(a, bd)
+                    else:
+                        raise HiYaPyCoImplementationException(
+                            'can not merge element from list of type %s to dict (@ "%s" try to merge "%s")' %
+                            (type(b), a, b,)
+                        )
+            else:
+                raise HiYaPyCoImplementationException(
+                    'can not merge %s to %s (@ "%s" try to merge "%s")' %
+                    (type(b), type(a), a, b,)
+                )
+        logger.debug('end deepmerge part: return: "%s"' % a)
+        logger.debug('<' * 30)
+        return a
+
     def _deepmerge(self, a, b):
         logger.debug('>'*30)
         logger.debug('deepmerge %s and %s' % (a, b,))
@@ -289,10 +407,11 @@ class HiYaPyCo():
                         if k in srcdicts.keys():
                             # we merge only if at least one key in dict is matching
                             merge = False
-                            for ak in ad.keys():
-                                if ak in srcdicts[k].keys():
-                                    merge = True
-                                    break
+                            if self.mergelists:
+                                for ak in ad.keys():
+                                    if ak in srcdicts[k].keys():
+                                        merge = True
+                                        break
                             if merge:
                                 logger.debug(
                                         'deepmerge ad: deep merge list dict elem w/ key:%s: "%s" and "%s"'
@@ -338,6 +457,9 @@ class HiYaPyCo():
         logger.debug('<'*30)
         return a
 
+    def yamlfiles(self):
+        return self._files
+
     def __str__(self):
         """String representation of the class"""
         return '%s [%s]' % (__name__, os.pathsep.join(self._files))
@@ -346,16 +468,16 @@ class HiYaPyCo():
         """return the data, merged and interpolated if required"""
         return self._data
 
-    def dump(self, default_flow_style = False):
+    def dump(self, **kwds):
         """dump the data as YAML"""
-        return dump(self._data, default_flow_style=default_flow_style)
+        return dump(self._data, **kwds)
 
-def dump(data, default_flow_style = False):
+def dump(data, **kwds):
     """dump the data as YAML"""
     if _usedefaultyamlloader:
-        return yaml.safe_dump(data, default_flow_style = default_flow_style)
+        return yaml.safe_dump(data, **kwds)
     else:
-        return odyldo.safe_dump(data, default_flow_style = default_flow_style)
+        return odyldo.safe_dump(data, **kwds)
 
 def load(*args, **kwargs):
     """
@@ -364,9 +486,12 @@ def load(*args, **kwargs):
 
     args: YAMLfile(s)
     kwargs:
-      * method: one of hiyapyco.METHOD_SIMPLE | hiyapyco.METHOD_MERGE
+      * method: one of hiyapyco.METHOD_SIMPLE | hiyapyco.METHOD_MERGE | hiyapyco.METHOD_SUBSTITUTE
+      * mergelists: boolean (default: True) try to merge lists (only makes sense if hiyapyco.METHOD_MERGE or hiyapyco.METHOD_SUBSTITUTE)
       * interpolate: boolean (default: False)
+      * castinterpolated: boolean (default: False) try to cast values after interpolating
       * usedefaultyamlloader: boolean (default: False)
+      * encoding: (default: 'utf-8') encoding used to read yaml files
       * loglevel: one of  the valid levels from the logging module
       * failonmissingfiles: boolean (default: True)
       * loglevelmissingfiles
